/* Mirror directrory structure to another host                   */
/*                                                               */
/* Usage: mirror <local_dir> <host> <user> <passwd> <remote_dir> */ 
/*                                                               */
/* Performs a mirror of the remote directory to the local using  */
/* the ftp protocol.  Sub directories are also mirrored.  Files  */
/* and directories are deleted on the local machine when required*/
/*                                                               */
/* Mirroring is based on timestamp and file size info            */
/*                                                               */
/* ideas for future enhancements :-                              */
/*                                                               */
/*   Read full perms of remote machine to mirror perms as well   */
/*   Read sym links and create locally                           */
/*   More efficient backup by moving files instead of deleting   */
/*     and then check the list of "moved" files before every get */
/*     (really del the temp directory before exiting)            */
/*   Delete hard linked files                                    */
/*   Verbose option                                              */
/*   Ability to mirror from more than one machine into one dir   */
/*                                                               */
/*                                                               */
/*   Peter Lord (plord@novell.co.uk)                             */
/*                                                               */
/*   (C) Novell 1995                                             */
/*                                                               */
/*   Based on an "uftp" - uses libftp                            */
/*                                                               */


#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <syslog.h>
#include <dirent.h>
#include <string.h>
#include <FtpLibrary.h>

#define TEMPDIR    "ls.out"

FTP *ftp;

main(int a,char **b)
{

#define LOCAL_DIR  b[1]
#define HOST       b[2]
#define USER       b[3]
#define PASSWD     b[4]
#define REMOTE_DIR b[5]

  if ( a < 6 )
    quit("Usage: mirror <local_dir> <host> <user> <passwd> <remote_dir>");

  
  FtplibDebug(yes);
  FtpLogin(&ftp,HOST,USER,PASSWD,NULL);
  FtpChdir(ftp,REMOTE_DIR);
  FtpBinary(ftp);
  doit(LOCAL_DIR);
  FtpBye(ftp);
  exit(0);
}

doit(char *dirname)
{
  DIR *dp;
  FILE *fp;
  struct dirent *de;
  struct tm tm, today;
  char n[256],fn[MAXNAMLEN];
  struct stat st;
  time_t k;
  
  int i, j, exists;

  int local_files, remote_files;
  char *local_names, *remote_names;
  off_t *local_size, *remote_size;
  time_t *local_mtime, *remote_mtime;
  mode_t *local_mode, *remote_mode;

  char mode[MAXNAMLEN], temp[MAXNAMLEN], month[MAXNAMLEN], time_year [MAXNAMLEN], name[MAXNAMLEN];
  int date, size;

  FtpDir (ftp, TEMPDIR );

  /* What is today ..*/
  k = time((time_t)0);
  gmtime_r (&k, &today);

  /* first, count the number of local files */
  if ( (dp=opendir(dirname)) == NULL )
  {
    log ("Cannot open local directory %s\n", dirname);
    return;
  }
  local_files = 0;
  while ( (de = readdir(dp)) != NULL )
    local_files++;
  closedir(dp);

  /* malloc the arrays */
  if ( (local_names = (char *) malloc ((local_files+1) * MAXNAMLEN)) == NULL)
    quit ("Cannot malloc\n");
  if ( (local_size = (off_t *) malloc ((local_files+1) * sizeof(off_t))) == NULL)
    quit ("Cannot malloc\n");
  if ( (local_mtime = (time_t *) malloc ((local_files+1) * sizeof(time_t))) == NULL)
    quit ("Cannot malloc\n");
  if ( (local_mode = (mode_t *) malloc ((local_files+1) * sizeof(mode_t))) == NULL)
    quit ("Cannot malloc\n");

  /* now build up the local filenames */

  if ( (dp=opendir(dirname)) == NULL )
  {
    log ("Cannot open local directory %s\n", dirname);
    return;
  }
  local_files = 0;
  while ( (de = readdir(dp)) != NULL )
  {
    if ( (strcmp (de -> d_name, ".") == 0) | (strcmp (de -> d_name, "..") == 0) )
      continue;
    sprintf(fn,"%s/%s",dirname,de->d_name);
    if ( stat(fn,&st) != 0 ) 
      log ( "cannot stat %s\n", de -> d_name);
    else
    {
      strncpy ( &local_names[local_files * MAXNAMLEN], de -> d_name, MAXNAMLEN);
      local_size[local_files] = st.st_size;
      local_mtime[local_files] = st.st_mtime;
      local_mode[local_files] = st.st_mode;
      local_files++;
    }
  }  
  closedir(dp);

  /* okay, now onto the remote system */

  if ( (fp = fopen (TEMPDIR, "r")) == NULL)
  {
    log ("Remote dir failed\n");
    return;
  }
  remote_files=0;
  fgets (temp, MAXNAMLEN, fp);
  while ( fgets (temp, MAXNAMLEN, fp) != NULL)
    remote_files++;
  fclose (fp);

  /* malloc the arrays */
  if ( (remote_names = (char *) malloc ((remote_files+1) * MAXNAMLEN)) == NULL)
    quit ("Cannot malloc\n");
  if ( (remote_size = (off_t *) malloc ((remote_files+1) * sizeof(off_t))) == NULL)
    quit ("Cannot malloc\n");
  if ( (remote_mtime = (time_t *) malloc ((remote_files+1) * sizeof(time_t))) == NULL)
    quit ("Cannot malloc\n");
  if ( (remote_mode = (mode_t *) malloc ((remote_files+1) * sizeof(mode_t))) == NULL)
    quit ("Cannot malloc\n");

  if ( (fp = fopen (TEMPDIR, "r")) == NULL)
  {
    log ("Remote dir failed\n");
    return;
  }
  remote_files=0;
  fgets (temp, MAXNAMLEN, fp);
  while ( (fgets (temp, MAXNAMLEN, fp) != NULL))
  {
    sscanf (temp, "%s %*s %*s %*s %d %s %d %s %s\n", &mode, &size, &month, &date, &time_year, &name);
    if ( (strcmp (name, ".") == 0) | (strcmp (name, "..") == 0) )
    {
      continue;
    }
    strncpy ( &remote_names[remote_files * MAXNAMLEN], name , MAXNAMLEN);
    sprintf (temp, "%02d %s %s", date, &month, &time_year);
    if (strchr (temp, ':') != NULL)
    {
      if (strptime (temp, "%d %b %R", &tm ) == NULL)
        log ("strptime failed\n");
      /* we don't know the year ... */
      if (tm.tm_mon <= today.tm_mon)
        tm.tm_year = today.tm_year;
      else
        tm.tm_year = today.tm_year - 1;
    }
    else
    {
      if (strptime (temp, "%d %b %Y", &tm ) == NULL)
        log ("strptime failed\n");
      /* this info is unknown */
      tm.tm_min = 0;
      tm.tm_hour = 0;
    }
    tm.tm_sec = 0;

    if (mode[0] == 'd')
      remote_mode[remote_files] = S_IFDIR;
    else
      if (mode[0] == 'l')
        remote_mode[remote_files] = S_IFLNK;
      else
        if (mode[0] == '-')
          remote_mode[remote_files] = S_IFREG;
        else
          remote_mode[remote_files] = 0;

    remote_size[remote_files] = size;
    remote_mtime[remote_files] = mktime (&tm);
    remote_files++;
  }

  /* debug lines */

  /*
  fprintf (stderr, "\nFound the following local files :-\n");
  for (i=0; i<local_files;i++)
    fprintf (stderr, "%s %ld %ld %ld\n", & local_names[i * MAXNAMLEN], local_size[i], local_mtime[i], local_mode[i]);

  fprintf (stderr, "\nFound the following remote files :-\n");
  for (i=0; i<remote_files;i++)
    fprintf (stderr, "%s %ld %ld %ld\n", & remote_names[i * MAXNAMLEN], remote_mtime[i], remote_size[i], remote_mode[i]);
  */


  /* okay, we can now work out what needs to be done */

  /* i)    If a filename in the remote list doesn't exist in the local, then
           copy the file */
  /* ii)   If the file name exists in both lists AND the size or timestamp
           differs, then copy the file */

  for (i=0; i<remote_files;i++)
  {
    if (remote_mode[i] & S_IFREG)
    {
      exists = 0;
      for (j=0; j<local_files; j++)
      {
        if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
        {
          exists = 1;
          if (remote_size[i] != local_size[j])
            fprintf (stderr, "Copy %s to %s/%s (size difference)\n", &remote_names[i * MAXNAMLEN], dirname, &remote_names[i * MAXNAMLEN]);
          else
            if (remote_mtime[i] != local_mtime[j])
              fprintf (stderr, "Copy %s to %s/%s (mtime difference)\n", &remote_names[i * MAXNAMLEN], dirname, &remote_names[i * MAXNAMLEN]);
        }
      }
      if (exists == 0)
        fprintf (stderr, "Copy %s to %s/%s (doesn't exist)\n", &remote_names[i * MAXNAMLEN], dirname, &remote_names[i * MAXNAMLEN]);
    }

  /* a)    If the remote filename is a directory AND the local directory
           exists then doit (file) */
  /* b)    If the remote filename is a directory AND the local directory
           doesn't exist then mkdir (file); then doit (file)  */
    else
    {
      if (remote_mode[i] & S_IFDIR)
      {
        exists = 0;
        for (j=0; j<local_files; j++)
          if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
            exists = 1;
        if (exists == 0)
          fprintf (stderr, "mkdir %s/%s (doesn't exist)\n", dirname, &remote_names[i * MAXNAMLEN]);
        fprintf (stderr, "FtpChdir(ftp,%s)\n", &remote_names[i * MAXNAMLEN]);
        fprintf (stderr, "doit (%s/%s)\n", dirname, &remote_names[i * MAXNAMLEN]);
        fprintf (stderr, "FtpChdir(ftp,..)\n");
      }
    }
  }

  /* iii)  If the file exists in the local list but doesn't on the remote,
           then delete */
  for (j=0; j<local_files;j++)
  {
    exists = 0;
    for (i=0; i<remote_files; i++)
    if ((strcmp(&remote_names[i * MAXNAMLEN], &local_names[j * MAXNAMLEN]) == 0))
      exists = 1;
    if (exists == 0)
      fprintf (stderr, "local file/dir %s must be deleted\n", &local_names[j * MAXNAMLEN]);
  }

  /* If a file is copied then it must be "touched" to make sure mtime is
   correct. */


  free (local_names);
  free (local_size);
  free (local_mtime);
  free (local_mode);
  free (remote_names);
  free (remote_size);
  free (remote_mtime);
  free (remote_mode);

  return (0);
}

quit(char *s)
{
  log(s);
  exit(1);
}


log (char *s)
{
  perror (s);
}
